from .eda import (dimensions,
describe_distribution,
identify_types,
num_to_categorical,
missing,
handling_missing,
drop_columns,
plot_categorical,
plot_numerical,
view_cardinality,
modify_cardinality,
correlation_pearson,
correlation_categorical,
correlation_phik,
num_variable_study,
interaction,
inspection,
split_and_scaling)

from .tools import (load_model, setup, scaling, ohe)


__all__ = [
    "dimensions",
    "describe_distribution",
    "identify_types",
    "num_to_categorical",
    "missing",
    "handling_missing",
    "drop_columns",
    "plot_categorical",
    "plot_numerical",
    "view_cardinality",
    "modify_cardinality",
    "correlation_pearson",
    "correlation_categorical",
    "correlation_phik",
    "num_variable_study",
    "interaction",
    "inspection",
    "split_and_scaling",
    "load_model",
    "setup",
    "scaling",
    "ohe",
]


import edamame.eda as eda
import pandas as pd
import numpy as np 
import matplotlib.pyplot as plt 
from typing import Union
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier
from xgboost import XGBClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC
from sklearn import datasets
from sklearn.metrics import confusion_matrix
import random
from typing import Tuple
import xgboost as xgb
from sklearn.metrics import RocCurveDisplay
from sklearn.metrics import roc_curve, auc


def check_random_forest(model: RandomForestClassifier) -> None: 
    """
    The function checks if the model passed is a random forest regression.

    Args:
        model (RandomForestClassifier): The input model to be checked.

    Raises:
            TypeError: If the input model is not a random forest regression model.
    
    Returns:
        None
    """
    if model.__class__.__name__ != 'RandomForestClassifier':
        raise TypeError('The model passed isn\'t a ridge model')
    

def check_xgboost(model: XGBClassifier) -> None: 
    """
    The function checks if the model passed is a xgboost regression.

    Args:
        model (xgb.XGBRegressor): The input model to be checked.

    Raises:
        TypeError: If the input model is not an XGBoost regression model.

    Returns:
        None
    """
    if model.__class__.__name__ != 'XGBClassifier':
        raise TypeError('The model passed isn\'t an xgboost')
    

class ClassifierDiagnose:
    """
    A class for diagnosing classification models.

    Attributes:
        X_train (pd.DataFrame): The input training data.
        y_train (pd.Series): The target training data.
        X_test (pd.DataFrame): The input test data.
        y_test (pd.Series): The target test data.

    Examples:
        >>> from edamame.classifier import TrainClassifier
        >>> classifier = TrainClassifier(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
        >>> nb = classifier.gaussian_nb()
        >>> classifiers_diagnose = ClassifierDiagnose(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
        >>> classifiers_diagnose.class_prediction_error(model=nb)
    """
    def __init__(self, X_train: pd.DataFrame, y_train: pd.DataFrame, X_test: pd.DataFrame, y_test: pd.DataFrame) -> None:
        self.X_train = X_train
        self.y_train = y_train
        self.X_test = X_test
        self.y_test = y_test

    
    def class_prediction_error(self, model: Union[LogisticRegression, GaussianNB, KNeighborsClassifier, DecisionTreeClassifier, RandomForestClassifier, XGBClassifier, SVC], train_data: bool = True) -> None:
        """
        This plot method shows the support (number of training samples) for each class in the fitted classification model as a stacked bar chart. Each bar is segmented to show the proportion of predictions (including false negatives and false positives, like a Confusion Matrix) for each class

        Args:
            model (Union[LogisticRegression, GaussianNB, KNeighborsClassifier, DecisionTreeClassifier, RandomForestClassifier, XGBClassifier, SVC]): Classification model.
            train_data (bool): Defines if you want to plot the stacked barplot on train or test data (train by default).
        
        Returns: 
            None
        """
        def stacked_barplot(matrix: np.array, num_of_class: int) -> None:
            cmap = plt.get_cmap('rainbow')
            plt.figure(figsize=(8,6))
            cm_bottom = np.zeros((num_of_class,))
            for i in range(len(matrix)):
                value = random.uniform(0, 1)
                clr = cmap(value)
                if i == 0:
                    plt.bar(range(num_of_class), matrix[i], label=f'Class {i}', color=clr)
                else: 
                    cm_bottom += matrix[i-1]
                    plt.bar(range(num_of_class), matrix[i], bottom=cm_bottom, label=f'Class {i}', color=clr)
            plt.xticks(range(num_of_class), [str(i) for i in range(num_of_class)])
            plt.ylabel('Count')
            plt.title('Predicted vs True Class')
            plt.legend()
            plt.show()

        if train_data:
            y_pred = model.predict(self.X_train)
            y_true = self.y_train.squeeze().to_numpy() 
            num_classes = y_true.max() + 1
            cm = confusion_matrix(y_true=y_true, y_pred=y_pred)
            stacked_barplot(matrix=cm, num_of_class=num_classes)
        else:
            y_pred = model.predict(self.X_test)
            y_true = self.y_test.squeeze().to_numpy() 
            num_classes = y_true.max() + 1
            cm = confusion_matrix(y_true=y_true, y_pred=y_pred)
            stacked_barplot(matrix=cm, num_of_class=num_classes)
    

    def random_forest_fi(self, model: RandomForestClassifier, figsize: Tuple[float, float] = (12,10)) -> None:
        """
        The method displays the feature importance plot of the random forest model. 

        Args:
            model (RandomForestClassifier): The input random forest model.

        Returns:
            None
        """
        check_random_forest(model)
        importances = model.feature_importances_
        std = np.std([tree.feature_importances_ for tree in model.estimators_], axis=0)
        feature_names = model.feature_names_in_
        forest_importances = pd.Series(importances, index=feature_names)
        plt.figure(figsize=figsize)
        forest_importances.plot.bar(yerr=std)
        plt.title("Feature importances using mean decrease in impurity")
        plt.ylabel("Mean decrease in impurity")
        plt.show()

    
    def xgboost_fi(self, model: XGBClassifier, figsize: tuple[float, float] = (14,12)) -> None:
        """
        The method displays the feature importance plot.

        Args:
            model (XGBClassifier): The input xgboost model.

        Returns:
            None
        """
        check_xgboost(model)
        xgb.plot_importance(model)
        plt.rcParams['figure.figsize'] = [figsize[0], figsize[1]]
        plt.show()

    
    def plot_roc_auc(self, model: Union[LogisticRegression, GaussianNB, KNeighborsClassifier, DecisionTreeClassifier, RandomForestClassifier, XGBClassifier, SVC], train_data: bool = True) -> None: 
        """
        Method for plotting the ROC curve and calculating the AUC values for a given model.

        Args: 
            model (Union[LogisticRegression, GaussianNB, KNeighborsClassifier, DecisionTreeClassifier, RandomForestClassifier, XGBClassifier, SVC]): Classification model.
            train_data (bool): Defines if you want to plot the stacked barplot on train or test data (train by default).
        
        Returns: 
            None
        """
        # One VS All strategy
        def _OVR_roc_curve(x, y, target_data):
            y_ohe = eda.ohe(y.iloc[:,0])
            y_score = model.predict_proba(x)
            n_classes = y_ohe.shape[1]
            fpr = dict()
            tpr = dict()
            roc_auc = dict()
            for i in range(n_classes):
                fpr[i], tpr[i], _ = roc_curve(y_ohe[:, i], y_score[:, i])
                roc_auc[i] = auc(fpr[i], tpr[i])
            # Plot ROC curve for each class
            plt.figure()
            colors = ['blue', 'red', 'green']
            for i, color in zip(range(n_classes), colors):
                plt.plot(fpr[i], tpr[i], color=color, lw=2, label=f'ROC curve of class {i} (area = {roc_auc[i]:.2f})')
            # Plot the random chance line
            plt.plot([0, 1], [0, 1], 'k--', lw=2)
            # Set plot properties
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.05])
            plt.xlabel('False Positive Rate')
            plt.ylabel('True Positive Rate')
            plt.title(f'One-vs-Rest multiclass ROC for {target_data}')
            plt.legend(loc="lower right")
            plt.show()
        # check binary problem: 
        if len(set(self.y_train.iloc[:,0])) > 2:
            if train_data:
                _OVR_roc_curve(self.X_train, self.y_train, target_data="Training data")
            else:
                _OVR_roc_curve(self.X_test, self.y_test, target_data="Test data")
        else:
            if train_data:
                roc = nb_roc = RocCurveDisplay.from_estimator(model, self.X_train, self.y_train)
                plt.show()
            else: 
                roc = nb_roc = RocCurveDisplay.from_estimator(model, self.X_test, self.y_test)
                plt.show()



import pickle
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, MinMaxScaler
import pandas as pd
from typing import Tuple
import numpy as np


def load_model(path: str):
    """
    The function load the model saved previously in the pickle format.

    Args:
        path (str): Path to the model saved in .pkl

    """
    with open(path, 'rb') as file:
        model = pickle.load(file)
    return model 


def setup(X: pd.DataFrame, y: pd.DataFrame, dummy: bool = False, seed: int = 42, size: float = 0.25) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    The function returns the following elements: X_train, y_train, X_test, y_test.

    Args: 
        X (pd.DataFrame): The model matrix X (features matrix).
        y (pd.DataFrame): The target variable.
        dummy (bool): If False, the function produces the OHE. If True, the dummy encoding.
        seed (int): Random seed to apply at the train_test_split function.
        size (float): Size of the test dataset. 

    Return:
        Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]: X_train, y_train, X_test, y_test.

    Example:
        >>> import edamame.eda as eda
        >>> df = pd.DataFrame({'category': ['A', 'B', 'A', 'B', 'C', 'A'], 'value': [1, 2, 3, 4, 4, 5], 'target': ['A2', 'A2', 'B2', 'B2', 'A2', 'B2']})
        >>> X, y = eda.split_and_scaling(df, 'target')
        >>> X_train, y_train, X_test, y_test = eda.setup(X, y)
    """
    # split dataset in train and test
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=size, random_state=seed)
    X_train = pd.get_dummies(data=X_train, drop_first=dummy)
    X_test = pd.get_dummies(data=X_test, drop_first=dummy)
    return X_train, y_train, X_test, y_test


def scaling(X: pd.DataFrame, minmaxscaler: bool = False) -> pd.DataFrame:
    """
    The function returns the normalised/standardized matrix.

    Args:
        X (pd.DataFrame): The model matrix X/X_train/X_test
    minmaxscaler (bool): Select the type of scaling to apply to the numerical columns. By default is setted to the StandardScaler. If minmaxscaler is set to True the numerical columns is trasfomed to [0,1].

    Return:
        pd.DataFrame

    Example:
        >>> import edamame.eda as eda
        >>> X = pd.DataFrame({'category': ['A', 'B', 'A', 'B', 'C', 'A'], 'value': [1, 2, 3, 4, 4, 5]})
        >>> X = eda.scaling(X)
    """
    # dataframe check 
    dataframe_review(X)
    # scaling quantitative variables 
    types = X.dtypes
    quant = types[types != 'object']
    quant_columns = list(quant.index)
    if minmaxscaler:
        scaler = MinMaxScaler()
    else: 
        scaler = StandardScaler()
    scaler.fit(X[quant_columns])
    X[quant_columns] = scaler.transform(X[quant_columns])
    return X


def ohe(array: np.ndarray) -> np.ndarray:
    """
    Convert a NumPy array that represents the categorical label of the target variable and transform it using one-hot encoding.

    Args:
        array (np.ndarray): The target variables passed in input.

    Return:
        np.ndarray: The one-hot encoded NumPy array.
    
    Example: 
        >>> import edamame.eda as eda
        >>> from sklearn import datasets
        >>> iris = datasets.load_iris()
        >>> y = iris.target
        >>> y_ohe = eda.ohe(y)
    """
    num_classes = len(np.unique(array))
    ohe_array = np.eye(num_classes)[array]
    return ohe_array


def dummy_control(data: pd.DataFrame) -> None:
    """
    The function checks if the Pandas dataframe passed is encoded with dummy or OHE. 
    
    Args:
        data (pd.Dataframe): A pandas DataFrame passed in input.

    Raises:
        TypeError: If the input DataFrame contains non-numerical columns.

    Returns:
        None
    """
    types = data.dtypes
    qual_col = types[types == 'object']
    if len(qual_col) != 0:
        raise TypeError('dataframe with non-numerical columns')
    else:
        pass


def dataframe_review(data: pd.DataFrame) -> None:
    """
    The function checks if the object passed is a Pandas dataframe.

    Args:
        data (pd.Dataframe): A pandas DataFrame passed in input.

    Raises:
        TypeError: If the input DataFrame contains non-numerical columns.

    Returns:
        None
    """
    if data.__class__.__name__ == 'DataFrame':
        pass
    else:
        raise TypeError('The data loaded is not a DataFrame')
    


import numpy as np
import pandas as pd
from  edamame.eda.tools import dataframe_review, dummy_control
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import StandardScaler
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier
from xgboost import XGBClassifier
from sklearn.model_selection import GridSearchCV, KFold, cross_val_score
from sklearn.metrics import confusion_matrix, classification_report
import seaborn as sns
import pickle
from IPython.display import display, Markdown
import matplotlib.pyplot as plt 
from typing import Tuple, Literal, List, Union
from sklearn.svm import SVC
# pandas options
pd.set_option('display.max_columns', None)
pd.set_option('display.width', None)
pd.set_option('display.max_colwidth', None)


class TrainClassifier:
    """
    This class represents a pipeline for training and handling classification models.

    Attributes: 
        X_train (pd.DataFrame): The input training data.
        y_train (pd.Series): The target training data.
        X_test (pd.DataFrame): The input test data.
        y_test (pd.Series): The target test data.
        
    Example:
        >>> from edamame.classifier import TrainClassifier
        >>> classifier = TrainClassifier(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
        >>> logistic = classifier.logistic()
        >>> classifier.model_metrics(model_name="logisitc")
        >>> classifier.model_save(model_name="logisitc")
        >>> nb = classifier.gaussian_nb()
        >>> knn = classifier.knn()
        >>> tree = classifier.tree()
        >>> rf = classifier.random_forest()
        >>> xgb = classifier.xgboost()
        >>> svm = classifier.svm()
        >>> classifier.model_metrics()
        >>> # using AutoML
        >>> models = classifier.auto_ml()
        >>> classifier.save_model()
    """
    def __init__(self, X_train, y_train, X_test, y_test):
        self.X_train = X_train
        self.y_train = y_train
        self.X_test = X_test
        self.y_test = y_test
        # check dataframe 
        dataframe_review(self.X_train)
        dataframe_review(self.X_test)
        # check columns type
        dummy_control(self.X_train)
        dummy_control(self.X_test)
        # init the model 
        self.__logistic_fit = {}
        self.__gaussian_nb_fit = {}
        self.__knn_fit = {}
        self.__tree_fit = {}
        self.__random_forest_fit = {}
        self.__xgb_fit = {}
        self.__svm_fit = {}


    def logistic(self, **kwargs) -> LogisticRegression:
        """
        Trains a logistic regression model using the training data and returns the fitted model.

        Args:
            **kwargs: Arbitrary keyword arguments to be passed to the `logistic` constructor.

        Returns:
            LogisticRegression: The trained logistic regression model.

        Example:
            >>> from edamame.classifier import TrainClassifier
            >>> classifier = TrainClassifier(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
            >>> logistic = classifier.logistic()
        """
        logistic = LogisticRegression(**kwargs)
        logistic.fit(self.X_train, self.y_train.squeeze())
        # save the model in the instance attributes
        self.__logistic_fit = logistic
        return self.__logistic_fit


    def gaussian_nb(self, **kwargs) -> GaussianNB:
        """
        Trains a Gaussian Naive Bayes classifier using the training data and returns the fitted model.

        Args:
            **kwargs: Arbitrary keyword arguments to be passed to the `Gaussian NB` constructor.

        Returns:
            GaussianNB: The trained Gaussian Naive Bayes classifier.
        
        Example:
            >>> from edamame.classifier import TrainClassifier
            >>> classifier = TrainClassifier(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
            >>> nb = classifier.gaussian_nb()
        """
        gauss_nb = GaussianNB(**kwargs)
        gauss_nb.fit(self.X_train, self.y_train.squeeze())
        # save the model in the instance attributes
        self.__gaussian_nb_fit = gauss_nb
        return self.__gaussian_nb_fit


    # ------------ #
    # KNN
    # ------------ #
    def knn(self, n_neighbors: Tuple[int, int, int] = (1, 50, 50), n_folds: int = 5, **kwargs) -> KNeighborsClassifier:
        """
        Train a k-Nearest Neighbors classification model using the training data, and perform a grid search to find the best value of 'n_neighbors' hyperparameter. 

        Args:
            n_neighbors (Tuple[int, int, int]): A tuple with three integers. The first and second integers are the range of the 'n_neighbors' hyperparameter that will be searched by the grid search, and the third integer is the number of values to generate in the interval [n_neighbors[0], n_neighbors[1]]. Default is [1, 50, 50].
            n_folds (int): The number of cross-validation folds to use for the grid search. Default is 5.
            **kwargs: Arbitrary keyword arguments to be passed to the `KNN` constructor.

        Returns:
            KNeighborsClassifier: The trained k-Nearest Neighbors classification model with the best 'n_neighbors' hyperparameter found by the grid search. 
        
        Example:
            >>> from edamame.classifier import TrainClassifier
            >>> classifier = TrainClassifier(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
            >>> knn = classifier.knn(n_neighbors=(1,50,50), n_folds=3) 
        """
        n_n = np.linspace(n_neighbors[0], n_neighbors[1], n_neighbors[2]).astype(np.int32)
        knn = KNeighborsClassifier(**kwargs)
        tuned_parameters = [{"n_neighbors": n_n}]
        grid_knn = GridSearchCV(knn, tuned_parameters, cv=n_folds, refit=True, verbose=0, scoring='accuracy')
        grid_knn.fit(self.X_train, self.y_train.squeeze())
        # save the model in the instance attributes
        self.__knn_fit = grid_knn.best_estimator_
        return self.__knn_fit


    def tree(self, alpha: Tuple[float, float, int] = (0., 0.001, 5), impurity: Tuple[float, float, int] = (0., 0.00001, 5), n_folds: int = 5, **kwargs) -> DecisionTreeClassifier:
        """
        Trains a decision tree classifier using the training data and returns the fitted model.
    
        Args:
            alpha (Tuple[float, float, int]): A tuple containing the minimum and maximum values of ccp_alpha and the number of values to try (default: (0., 0.001, 5)).
            impurity (Tuple[float, float, int]): A tuple containing the minimum and maximum values of min_impurity_decrease and the number of values to try (default: (0., 0.00001, 5)).
            n_folds (int): The number of cross-validation folds to use for grid search (default: 5).
            **kwargs: Arbitrary keyword arguments to be passed to the `tree` constructor.

        Returns:
            DecisionTreeClassifier: The trained decision tree classifier model.

        Example:
            >>> from edamame.classifier import TrainClassifier
            >>> classifier = TrainClassifier(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
            >>> tree = classifier.tree(alpha=(0., 0.001, 5), impurity=(0., 0.00001, 5), n_folds=3) 
        """
        alphas = np.linspace(alpha[0], alpha[1], alpha[2])
        impurities = np.linspace(impurity[0], impurity[1], impurity[2])
        tuned_parameters = [{"ccp_alpha": alphas, 'min_impurity_decrease': impurities}]
        tree = DecisionTreeClassifier(**kwargs) 
        grid_tree = GridSearchCV(tree, tuned_parameters, cv=n_folds, refit=True, verbose=0, scoring='accuracy')
        grid_tree.fit(self.X_train, self.y_train.squeeze())
        # save the model in the instance attributes
        self.__tree_fit = grid_tree.best_estimator_
        return self.__tree_fit


    def random_forest(self, n_estimators: Tuple[int, int, int] = (50, 1000, 5), n_folds: int = 2, **kwargs) -> RandomForestClassifier:
        """
        Train a Random Forest classifier using the training data and return the fitted model.

        Args:
            n_estimators (Tuple[int, int, int]): The range of the number of trees in the forest. Default is (50, 1000, 5).
            n_folds (int): The number of folds in cross-validation. Default is 2.
            **kwargs: Arbitrary keyword arguments to be passed to the `random forest` constructor.

        Returns:
            RandomForestClassifier: The trained Random Forest classifier.

        Example:
            >>> from edamame.classifier import TrainClassifier
            >>> classifier = TrainClassifier(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
            >>> rf = classifier.random_forest(n_estimators=(50, 1000, 5), n_folds=2) 
        """
        n_estimators = np.linspace(n_estimators[0], n_estimators[1], n_estimators[2]).astype(np.int16)
        tuned_parameters = [{"n_estimators": n_estimators}]
        random_forest = RandomForestClassifier(warm_start=True, n_jobs=-1, **kwargs)
        grid_random_forest = GridSearchCV(random_forest, tuned_parameters, cv=n_folds, refit=True, verbose=0, scoring='accuracy')
        grid_random_forest.fit(self.X_train, self.y_train.squeeze())
        # save the model in the instance attributes
        self.__random_forest_fit = grid_random_forest.best_estimator_
        return self.__random_forest_fit


    def xgboost(self, n_estimators: Tuple[int, int, int] = (10, 100, 5), n_folds: int = 2, **kwargs) -> XGBClassifier:
        """
        Train an XGBoost classifier using the training data and return the fitted model.

        Args:
            n_estimators (Tuple[int, int, int]): The range of the number of boosting rounds. Default is (10, 100, 5).
            n_folds (int): The number of folds in cross-validation. Default is 2.
            **kwargs: Arbitrary keyword arguments to be passed to the `xgboost` constructor.

        Returns:
            XGBClassifier: The trained XGBoost classifier.

        Example:
            >>> from edamame.classifier import TrainClassifier
            >>> classifier = TrainClassifier(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
            >>> xgboost = classifier.xgboost(n_estimators=(10, 100, 5), n_folds=2) 
        """
        n_est = np.linspace(n_estimators[0], n_estimators[1], n_estimators[2]).astype(np.int16)
        tuned_parameters = {"n_estimators": n_est}
        xgb_m = XGBClassifier(**kwargs)
        grid_xgb = GridSearchCV(xgb_m, tuned_parameters, cv=n_folds, refit=True, verbose=0, scoring='accuracy')
        grid_xgb.fit(self.X_train, self.y_train.squeeze())
        # save the model in the instance attributes
        self.__xgb_fit = grid_xgb.best_estimator_
        return self.__xgb_fit
    

    def svm(self, n_folds: int = 2, *args, **kwargs) -> SVC:
        """
        Trains an SVM classifier using the training data and returns the fitted model.

        Args:
            n_folds (int): The number of folds in cross-validation. Default is 2.
            *args: Variable length argument list to be passed to the `SVC` constructor.
            **kwargs: Arbitrary keyword arguments to be passed to the `SVC` constructor.
        
        Returns:
            SVC: The trained SVM classifier.
        
        Example: 
            >>> from edamame.classifier import TrainClassifier
            >>> classifier = TrainClassifier(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
            >>> svm = classifier.svm(kernel="linear", C=1.0, gamma="auto")
        """
        n_kernel = ["linear", "poly", "rbf", "sigmoid"] 
        tuned_parameters = {"kernel": n_kernel}
        svm_c = SVC(*args, **kwargs)
        grid_svm_c = GridSearchCV(svm_c, tuned_parameters, cv=n_folds, refit=True, verbose=0, scoring='accuracy')
        grid_svm_c.fit(self.X_train, self.y_train.squeeze())
        # save the model in the instance attributes
        self.__svm_fit = grid_svm_c.best_estimator_
        return self.__svm_fit


    def model_metrics(self, model_name: Literal["all", "logistic", "guassian_nb", "knn", "tree", "random_forest", "xgboost", "svm"] = 'all', cm: bool = False) -> None:
        """
        Display classification metrics (confusion matrix and classification report) for specified or all trained models.

        Args:
            model_name (Literal["all", "logistic", "guassian_nb", "knn", "tree", "random_forest", "xgboost", "svm"]): The name of the model to display the metrics for. Defaults to 'all'.
            cm (bool): Whether to display the confusion matrix. Defaults to False.

        Returns:
            None

        Example:
            >>> from edamame.classifier import TrainClassifier
            >>> classifier = TrainClassifier(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
            >>> xgboost = classifier.xgboost(n_estimators=(10, 100, 5), n_folds=2) 
            >>> classifier.model_metrics(model_name="xgboost")
        """
        model_dct = {'logistic': 0, 'guassian_nb': 1, 'knn': 2, 'tree': 3, 'random_forest': 4, 'xgboost': 5, 'svm': 6}
        model_list = [self.__logistic_fit, self.__gaussian_nb_fit, self.__knn_fit, self.__tree_fit, self.__random_forest_fit, self.__xgb_fit, self.__svm_fit]
        if model_name == 'all':
            for key in model_dct:
                if model_list[model_dct[key]].__class__.__name__ == 'dict':
                        display(f'unable to show {key} model metrics')
                else:
                    title = f'### {key} model metrics:'
                    display(Markdown(title))
                    y_pred_train = model_list[model_dct[key]].predict(self.X_train)
                    y_pred_test = model_list[model_dct[key]].predict(self.X_test)
                    if confusion_matrix:
                        plt.figure(figsize=(10,4))
                        plt.subplot(121)
                        sns.heatmap(confusion_matrix(self.y_train, y_pred_train), annot=True, fmt="2.0f")
                        plt.title(f'{key} train')
                        plt.subplot(122)
                        sns.heatmap(confusion_matrix(self.y_test, y_pred_test), annot=True, fmt="2.0f")
                        plt.title(f'{key} test')
                        plt.show()
                    title = f'#### Train classification report'
                    display(Markdown(title))
                    print(classification_report(self.y_train, y_pred_train))
                    title = f'#### Test classification report'
                    display(Markdown(title))
                    print(classification_report(self.y_test, y_pred_test))
        else:
            if model_list[model_dct[model_name]].__class__.__name__ == 'dict':
                display(f'unable to show {model_name} model metrics')
            else:
                title = f'### {model_name} model metrics:'
                display(Markdown(title))
                y_pred_train = model_list[model_dct[model_name]].predict(self.X_train)
                y_pred_test = model_list[model_dct[model_name]].predict(self.X_test)
                if confusion_matrix: 
                    plt.figure(figsize=(10,4))
                    plt.subplot(121)
                    sns.heatmap(confusion_matrix(self.y_train, y_pred_train), annot=True, fmt="2.0f")
                    plt.title(f'{model_name} train')
                    plt.subplot(122)
                    sns.heatmap(confusion_matrix(self.y_test, y_pred_test), annot=True, fmt="2.0f")
                    plt.title(f'{model_name} test')
                    plt.show()
                title = f'#### Train classification report'
                display(Markdown(title))
                print(classification_report(self.y_train, y_pred_train))
                title = f'#### Test classification report'
                display(Markdown(title))
                print(classification_report(self.y_test, y_pred_test))


    def auto_ml(self, n_folds: int = 5, data: Literal['train', 'test'] = 'train') -> List:
        """
        Perform automated machine learning with cross validation on a list of classification models.
        
        Args:
            n_folds (int): Number of cross-validation folds. Defaults to 5.
            data (Literal['train', 'test']): Target dataset for cross-validation. 
                Must be either 'train' or 'test'. Defaults to 'train'.
        
        Returns:
            List: List of best-fit classification models for each algorithm.

        Example:
            >>> from edamame.classifier import TrainClassifier
            >>> classifier = TrainClassifier(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
            >>> model_list = classifier.auto_ml()
        """
        kfold = KFold(n_splits=n_folds)
        cv_mean = []
        score = []
        std = []
        classifier = ["Logistic", "Gaussian NB", "KNN", "Tree", "Random forest", "Xgboost", "SVM"]
        try:
            model_list = [LogisticRegression(), GaussianNB(), KNeighborsClassifier(n_neighbors=self.__knn_fit.n_neighbors),
                          DecisionTreeClassifier(ccp_alpha=self.__tree_fit.ccp_alpha, min_impurity_decrease=self.__tree_fit.min_impurity_decrease),
                          RandomForestClassifier(n_estimators=self.__random_forest_fit.n_estimators, warm_start=True, n_jobs=-1), 
                          XGBClassifier(n_estimators=self.__xgb_fit.n_estimators),
                          SVC(kernel=self.__svm_fit.kernel)]
        except:
            # find best hyperparameters
            self.logistic()
            self.gaussian_nb()
            self.knn()
            self.tree()
            self.random_forest()
            self.xgboost()
            self.svm()
            # model list 
            model_list = [LogisticRegression(), GaussianNB(), KNeighborsClassifier(n_neighbors=self.__knn_fit.n_neighbors),
                          DecisionTreeClassifier(ccp_alpha=self.__tree_fit.ccp_alpha, min_impurity_decrease=self.__tree_fit.min_impurity_decrease),
                          RandomForestClassifier(n_estimators = self.__random_forest_fit.n_estimators, warm_start=True, n_jobs=-1),
                          XGBClassifier(n_estimators = self.__xgb_fit.n_estimators),
                          SVC(kernel=self.__svm_fit.kernel)]
        # cross validation loop 
        for model in model_list:
            if data == 'train':
                cv_result = cross_val_score(model, self.X_train, self.y_train.squeeze(), cv=kfold, scoring="accuracy")
            elif data == 'test':
                cv_result = cross_val_score(model, self.X_test, self.y_test.squeeze(), cv=kfold, scoring="accuracy")
            else:
                raise ValueError('insert valid target dataset (\'train\' or \'test\')')
            cv_mean.append(cv_result.mean())
            std.append(cv_result.std())
            score.append(cv_result)
        # dataframe for results 
        df_kfold_result = pd.DataFrame({"CV Mean": cv_mean, "Std": std}, index=classifier)
        # display step 
        string = f'### Metrics results on {data} set:'
        display(Markdown(string))
        display(df_kfold_result)
        # boxplot on R2
        box = pd.DataFrame(score, index=classifier)
        plt.figure(figsize=(10,8))
        box.T.boxplot()
        plt.show()

        return [self.__logistic_fit, self.__gaussian_nb_fit, self.__knn_fit, self.__tree_fit, self.__random_forest_fit, self.__xgb_fit, self.__svm_fit]


    def save_model(self, model_name: Literal["all", "logistic", "guassian_nb", "knn", "tree", "random_forest", "xgboost", "svm"] = 'all') -> None:
        """
        Saves the specified machine learning model or all models in the instance to a pickle file.

        Args:
            model_name (Literal["all", "linear", "lasso", "ridge", "tree", "random_forest", "xgboost", "svm"]): 
                The name of the model to save. Defaults to 'all'.
            
        Returns:
            None

        Example:
            >>> from edamame.classifier import TrainClassifier
            >>> classifier = TrainClassifier(X_train=X_train, y_train=y_train, X_test=X_test, y_test=y_test)
            >>> model_list = classifier.auto_ml()
            >>> classifier.save_model(model_name="all")
        """
        model_dct = {'logistic': 0, 'guassian_nb': 1, 'knn': 2, 'tree': 3, 'random_forest': 4, 'xgboost': 5, 'svm': 6}
        model_list = [self.__logistic_fit, self.__gaussian_nb_fit, self.__knn_fit, self.__tree_fit, self.__random_forest_fit, self.__xgb_fit, self.__svm_fit]
        if model_name == 'all':
            for key in model_dct:
                if model_list[model_dct[key]].__class__.__name__ == 'dict':
                        display(f'unable to save {key} model')
                else:
                    filename = f'{key}.pkl'
                    with open(filename, 'wb') as file:
                        pickle.dump(model_list[model_dct[key]], file)
                        display(f'{filename} saved')
        else:
            if model_list[model_dct[model_name]].__class__.__name__ == 'dict':
                display(f'unable to save {model_name} model')
            else:
                filename = f'{model_name}.pkl'
                with open(filename, 'wb') as file:
                    pickle.dump(model_list[model_dct[model_name]], file)



def classifier_metrics(model: Union[LogisticRegression, GaussianNB, KNeighborsClassifier, DecisionTreeClassifier, RandomForestClassifier, XGBClassifier, SVC], X: pd.DataFrame, y: pd.DataFrame) -> None:
    """
    Display classification metrics (confusion matrix and classification report) for the model passed as input to the function.

    Args:
        model (Union[LogisticRegression, GaussianNB, KNeighborsClassifier, DecisionTreeClassifier, RandomForestClassifier, XGBClassifier, SVC]): Classification model.
        X (pd.DataFrame): Input features.
        y (pd.DataFrame): Target feature.
    
    Returns:
        None
    """
    dataframe_review(X)
    dummy_control(X)
    y_pred = model.predict(X)
    title = f'#### Model metrics:'
    display(Markdown(title))
    sns.heatmap(confusion_matrix(y, y_pred), annot=True, fmt="2.0f")
    plt.show()
    print(classification_report(y, y_pred))

        
